/*
PxDbgC - Ptokax Debug Console Window
Copyright (C) 2005-2006  Morten Larsen, bluebear dot dk at gmail dot com, http://www.thewildplace.dk/

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-------------------------------------------------------------------------------
Whats new (v0.3):
1. function to close the console window added (requested by bastya_elvtars)
2. function to open the console window added, must always be call before the console window shows
3. Disable close of the console from UI (requested by bastya_elvtars)
4. Lib is now released under the GPL license; means its open source now :o)

Changes (v0.4):
1. Compiled as C++ rather then C code
2. Some code beautifies still some to go
3. Build for Lua 5.1 - Lua 5.0.2 support in further development will not be supported anymore.

ToDo@ Compact code, make the print fuctions into one, however this is no rush and im lazy ;)
*/

#include "windows.h"
#include "stdio.h"

extern "C" {
#ifdef _LUA5
    #include "..\lua5\include\lua.h"
    #include "..\lua5\include\lualib.h"
    #include "..\lua5\include\lauxlib.h"

    #pragma link "../lua5/PXLua.lib"
    #pragma message("Linking lua5 PXLua.lib")
#else
    #include "..\lua51\src\lua.h"
    #include "..\lua51\src\lualib.h"
    #include "..\lua51\src\lauxlib.h"

    #pragma link "../lua51/PXLua.lib"
    #pragma message("Linking lua51 PXLua.lib")
#endif
}

const char *consoletitle = "PxDbgC.dll 0.4 [Lua 5.1] - by bluebear (http://www.thewildplace.dk/) - Ptokax Script Debug Window";

HANDLE hstdout = NULL;
int ontop = 0;
char *logfile = NULL;
FILE *lfile;

// Yes i know the func name is badly choosen.. originally it had the fmt, ... but thats not needed here so was removed and name keept
#define printf __Printf
int __Printf(lua_State *L, const char *str)
{
  DWORD ccharswritten;
  if(hstdout) {
	WriteConsole(hstdout, str, strlen(str), &ccharswritten, NULL);
  }

  if (logfile) {
	lfile = fopen( logfile, "a" );
	if (lfile == NULL) {
	  lua_pushliteral(L, "OS reported error while trying to write logfile!");
	  lua_error(L);
	  return 0;
	}
	fwrite(str, sizeof( char ), strlen(str), lfile);
	fclose(lfile);
  }
  return 0;
}

int lua_stoplog (lua_State *L) {
  /* Arg checking here is not really needed
	 but we do it anyway */
  int argc = lua_gettop(L);
  if (argc != 0) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called without arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }

  if (logfile) {
	free(logfile);
	logfile = NULL;
  }
  return 0;
}

int lua_startlog (lua_State *L) {
  int argc = lua_gettop(L);
  if (argc != 1) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called with 1 argument!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }
  if (lua_isstring(L, 1) == FALSE) {
	luaL_typerror(L, 1, "String");
	return 0;
  }

  /* If log already runs delte filename so we wont have a leak */
  if ( logfile ) {
	free(logfile);
  }

  logfile = (char *) malloc (lua_strlen(L, 1));
  if (logfile == NULL) {
	lua_pushliteral(L, "Out of memory!");
	lua_error(L);
	return 0;
  }

  strcpy(logfile, lua_tostring(L, 1));

  lfile = fopen(logfile, "w"); /* create file */
  if (lfile == NULL) {
	lua_pushliteral(L, "OS reported error while trying to create logfile!");
	lua_error(L);
	return 0;
  }
  fclose (lfile);

  return 0;
}

// Todo@ Make next 3 fucn's code into 1
int lua_print (lua_State *L) {
  int i;
  int argc = lua_gettop(L);
  if (argc < 1) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called with 1 or more arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }
  for (i=1; i<=argc; i++) {
	if (lua_isstring(L, i) == FALSE) {
	  luaL_typerror(L, i, "String");
	  return 0;
	}
	printf(L, lua_tostring(L, i));
  } // for
  return 0;
}

int lua_println (lua_State *L) {
  int i;
  int argc = lua_gettop(L);
  if (argc < 1) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called with 1 or more arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }
  for (i=1; i<=argc; i++) {
	if (lua_isstring(L, i) == FALSE) {
	  luaL_typerror(L, i, "String");
	  return 0;
	}
	printf(L, lua_tostring(L, i));
  } // for
  printf(L, "\n");
  return 0;
}

int lua_printlns (lua_State *L) {
  int i;
  int argc = lua_gettop(L);
  if (argc < 1) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called with 1 or more arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }
  for (i=1; i<=argc; i++) {
	if (lua_isstring(L, i) == FALSE) {
	  luaL_typerror(L, i, "String");
	  return 0;
	}
	printf(L, lua_tostring(L, i));
	printf(L, "\n");
  } // for
  return 0;
}

int lua_msgbox(lua_State* L) {
  const char* message;
  const char* caption;
  int type;
  int i;
  int result;
  int argc = lua_gettop(L);
  if (argc != 3) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called with 3 arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }
  for ( i=1; i<=2; i++ )
  {
	if (lua_isstring(L, i) == FALSE) {
	  luaL_typerror(L, i, "String");
	  return 0;
	}
  }
  if (lua_isnumber(L, 3) == FALSE) {
	luaL_typerror(L, 3, "Integral");
	return 0;
  }

  message = lua_tostring(L, 1);
  caption = luaL_optstring(L, 2, "PxDbgC");
  type = lua_tonumber(L, 3);
  if (type == 0) {
	result = MessageBox(NULL, message, caption, MB_OK);
  } else if (type == 1) {
	result = MessageBox(NULL, message, caption, MB_YESNO);
  } else {
	luaL_argerror( L, 3, "Messagebox type (valid values is 0 (zero) or 1.");
	return 0;
  }

  if (logfile) {
	lfile = fopen(logfile, "a");
	if (lfile == NULL) {
	  lua_pushliteral(L, "OS reported error while trying to write logfile!");
	  lua_error(L);
	  return 0;
	}

	if (type == 0) {
	  fwrite("[msgbox] OK M->", sizeof( char ), 15, lfile);
	} else {
	  fwrite("[msgbox] YESNO M->", sizeof( char ), 18, lfile);
	}
	fwrite(message, sizeof( char ), strlen(message), lfile);
	fwrite(" C->", sizeof( char ), 4, lfile);
	fwrite(caption, sizeof( char ), strlen(caption), lfile);
	if (result == 6) {
	  fwrite(" R->YES", sizeof( char ), 7, lfile);
	} else if (result == 7) {
	  fwrite(" R->NO", sizeof( char ), 6, lfile);
	} else {
	  fwrite(" R->OK", sizeof( char ), 6, lfile);
	}
	fwrite("\n", sizeof( char ), 1, lfile);
	fclose(lfile);
  }

  lua_pushnumber(L, result);
  return 1;
}

int lua_always_ontop (lua_State *L) {
  HWND cnsl;
  RECT rccnsl;

  int argc = lua_gettop(L);
  if ( argc != 0 ) {
	luaL_where(L, 1);
	lua_pushliteral(L, "Function should be called without arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }

  if ((ontop == 0) && (hstdout)) {
	ontop = 1;
	cnsl = GetConsoleWindow();
	GetWindowRect(cnsl, &rccnsl);
	SetWindowPos(cnsl, HWND_TOPMOST, rccnsl.right, rccnsl.bottom , 0, 0, SWP_NOSIZE);
  }
  return 0;
}

/* Disable close of the console from UI
	 requested by bastya_elvtars */
void noclose (void) {
  HMENU hmenu;
  HWND hwnd = NULL;
  while ( NULL == hwnd ) { /* We give it a few trys; dont always work 1st time around */
	hwnd = FindWindowEx (NULL, NULL, NULL, (LPCTSTR)consoletitle);
  }
  hmenu = GetSystemMenu (hwnd, FALSE);
  DeleteMenu (hmenu, SC_CLOSE, MF_BYCOMMAND);
  return;
}

int lua_opencnsl (lua_State *L) {
  if (hstdout) {
	lua_pushnil(L);
	return 1;
  }
  // FreeConsole(); //patch eventually for Px beta with lua 5.1
  if (AllocConsole() != 0) {
	hstdout = GetStdHandle(STD_OUTPUT_HANDLE);
	SetConsoleTitle(consoletitle);
	noclose(); /* Disables the close button */

	printf(L, "PxDbgC - Ptokax Debug Console Window\n");
	printf(L, "Copyright (C) 2005-2006  Morten Larsen, bluebear dot dk at gmail dot com, http://www.thewildplace.dk/\n\n");
	printf(L, "This program is free software; you can redistribute it and/or\n");
	printf(L, "modify it under the terms of the GNU General Public License\n");
	printf(L, "as published by the Free Software Foundation; either version 2\n");
	printf(L, "of the License, or (at your option) any later version.\n\n");
	printf(L, "This program is distributed in the hope that it will be useful,\n");
	printf(L, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	printf(L, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	printf(L, "GNU General Public License for more details.\n\n");
	printf(L, "You should have received a copy of the GNU General Public License\n");
	printf(L, "along with this program; if not, write to the Free Software\n");
	printf(L, "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\n");
	printf(L, "--------------------------------------------------------------\n\n");
	lua_pushnil(L);
  } else {
	lua_pushstring(L, "Could not open console window!");
  }

  return 1;
}

/* Requested by bastya_elvtars
   function to close the console window */
int lua_closecnsl (lua_State *L) {
  int argc = lua_gettop(L);
  if (argc != 0) {
	luaL_where(L,1);
	lua_pushliteral(L, "Function should be called without arguments!");
	lua_concat(L, 2);
	lua_error(L);
	return 0;
  }

  if (!hstdout) {
	lua_pushnil(L);
	return 1;
  }

  if (FreeConsole() != 0) {
	hstdout = NULL;
	ontop = 0;
	lua_pushnil(L);
  } else {
	lua_pushstring(L, "Could not close console window!");
  }

  return 1;
}

static const luaL_reg console_methods[] = {
		{"Open", lua_opencnsl},
		{"Close", lua_closecnsl},
		{"AlwaysOnTop", lua_always_ontop},
		{"StopLog", lua_stoplog},
		{"StartLog", lua_startlog},
		{"PrintLns", lua_printlns},
		{"PrintLn", lua_println},
		{"Print", lua_print},
		{"MsgBox", lua_msgbox},
		{0,0}
};

#ifdef _LUA5
#define newtable(L) (lua_newtable(L), lua_gettop(L))
#endif

int console_register (lua_State *L) {
#ifdef _LUA5
  int metatable, methods;
  lua_pushliteral(L, "DbgC");
  methods   = newtable(L);
  metatable = newtable(L);
  lua_pushliteral(L, "__index");
  lua_pushvalue(L, methods);
  lua_settable(L, metatable);
  lua_pushliteral(L, "__metatable");
  lua_pushvalue(L, methods);
  lua_settable(L, metatable);
  luaL_openlib(L, 0, console_methods, 1);
  lua_settable(L, LUA_GLOBALSINDEX);
#else
  luaL_register(L, "DbgC", console_methods); // We only need this for Lua 5.1
#endif
  return 0;
}

// Register lib functions (make useable in script)
int __declspec(dllexport) libinit (lua_State* L) {
  console_register(L);
  return 0;
}
